/* Copyright (C) 2001-2003 Peter Selinger.
   This file is part of potrace. It is free software and it is covered
   by the GNU general public license. See the file COPYING for details. */

/* $Id: backend_svg.c,v 1.6 2003/08/10 23:02:48 selinger Exp $ */

/* The Postscript backend of potrace. This can produce "ps" or "svg"
   output, in ASCII or binary format, and different kinds of graphical
   debugging output. Postscript compression is optionally supplied via
   the functions in flate.c. */

#include <stdio.h>
#include <stdarg.h>
#include <string.h>
#include <math.h>

#include "main.h"
#include "backend_svg.h"
#include "curve.h"
#include "lists.h"

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

/* ---------------------------------------------------------------------- */
/* path-drawing auxiliary functions */

#define unit(x) (int)(floor(x*info.unit+.5))

static long curx, cury;
static char lastop = 0;
static int column = 0;
static int newline = 1;

static void shiptoken(FILE *fout, char *token) {
  int c = strlen(token);
  if (!newline && column+c+1 > 75) {
    fprintf(fout, "\n");
    column = 0;
    newline = 1;
  } else if (!newline) {
    fprintf(fout, " ");
    column++;
  }
  fprintf(fout, "%s", token);
  column += c;
  newline = 0;
}

static void ship(FILE *fout, char *fmt, ...) {
  va_list args;
  static char buf[4096]; /* static string limit is okay here because
			    we only use constant format strings */
  char *p, *q;

  va_start(args, fmt);
  vsnprintf(buf, 4095, fmt, args);
  buf[4095] = 0;
  va_end(args);

  p = buf;
  while ((q = strchr(p, ' ')) != NULL) {
    *q = 0;
    shiptoken(fout, p);
    p = q+1;
  }
  shiptoken(fout, p);
}

static void svg_moveto(FILE *fout, dpoint_t p) {
  curx = unit(p.x);
  cury = unit(p.y);

  ship(fout, "M%ld %ld", curx, cury);
  lastop = 'M';
}

static void svg_rmoveto(FILE *fout, dpoint_t p) {
  long xx, yy;

  xx = unit(p.x);
  yy = unit(p.y);
  
  ship(fout, "m%ld %ld", xx-curx, yy-cury);

  curx = xx;
  cury = yy;
  lastop = 'm';
}

static void svg_lineto(FILE *fout, dpoint_t p) {
  long xx, yy;

  xx = unit(p.x);
  yy = unit(p.y);

  if (lastop != 'l') {
    ship(fout, "l%ld %ld", xx-curx, yy-cury);
  } else {
    ship(fout, "%ld %ld", xx-curx, yy-cury);
  }
  curx = xx;
  cury = yy;
  lastop = 'l';
}

static void svg_curveto(FILE *fout, dpoint_t p1, dpoint_t p2, dpoint_t p3) {
  long xx1, yy1, xx2, yy2, xx3, yy3;

  xx1 = unit(p1.x);
  yy1 = unit(p1.y);
  xx2 = unit(p2.x);
  yy2 = unit(p2.y);
  xx3 = unit(p3.x);
  yy3 = unit(p3.y);

  if (lastop != 'c') {
    ship(fout, "c%ld %ld %ld %ld %ld %ld", xx1-curx, yy1-cury, xx2-curx, yy2-cury, xx3-curx, yy3-cury);
  } else {
    ship(fout, "%ld %ld %ld %ld %ld %ld", xx1-curx, yy1-cury, xx2-curx, yy2-cury, xx3-curx, yy3-cury);
  }
  curx = xx3;
  cury = yy3;
  lastop = 'c';
}

/* ---------------------------------------------------------------------- */
/* functions for converting a path to an SVG path element */

/* Explicit encoding. If abs is set, move to first coordinate
   absolutely. */
static int svg_path(FILE *fout, curve_t *curve, int m, int abs) {
  int i;
  dpoint_t *c;

  c = curve[m-1].c;
  if (abs) {
    svg_moveto(fout, c[2]);
  } else {
    svg_rmoveto(fout, c[2]);
  }

  for (i=0; i<m; i++) {
    c = curve[i].c;
    switch (curve[i].tag) {
    case CORNER:
      svg_lineto(fout, c[1]);
      svg_lineto(fout, c[2]);
      break;
    case CURVETO:
      svg_curveto(fout, c[0], c[1], c[2]);
      break;
    }
  }
  newline = 1;
  shiptoken(fout, "z");
  return 0;
}

/* produce a jaggy path - for debugging. If abs is set, move to first
   coordinate absolutely. If abs is not set, move to first coordinate
   relatively, and traverse path in the opposite direction. */
static int svg_jaggy_path(FILE *fout, point_t *pt, int n, int abs) {
  int i;
  point_t cur, prev;
  
  if (abs) {
    cur = prev = pt[n-1];
    svg_moveto(fout, dpoint(cur));
    for (i=0; i<n; i++) {
      if (pt[i].x != cur.x && pt[i].y != cur.y) {
	cur = prev;
	svg_lineto(fout, dpoint(cur));
      }
      prev = pt[i];
    }
    svg_lineto(fout, dpoint(pt[n-1]));
  } else {
    cur = prev = pt[0];
    svg_rmoveto(fout, dpoint(cur));
    for (i=n-1; i>=0; i--) {
      if (pt[i].x != cur.x && pt[i].y != cur.y) {
	cur = prev;
	svg_lineto(fout, dpoint(cur));
      }
      prev = pt[i];
    }
    svg_lineto(fout, dpoint(pt[0]));
  }
  newline = 1;
  shiptoken(fout, "z");
  return 0;
}

/* ---------------------------------------------------------------------- */
/* Backend. */

/* public interface for SVG */
int page_svg(FILE *fout, path_t *plist, imginfo_t *imginfo) {
  path_t *p;

  int bboxx = ceil(imginfo->trans.bb[0]+imginfo->lmar+imginfo->rmar);
  int bboxy = ceil(imginfo->trans.bb[1]+imginfo->tmar+imginfo->bmar);
  double origx = imginfo->trans.orig[0] + imginfo->lmar;
  double origy = bboxy - imginfo->trans.orig[1] - imginfo->bmar;
  double scalex = imginfo->width / imginfo->pixwidth / info.unit;
  double scaley = -imginfo->height / imginfo->pixheight / info.unit;

  /* header */
  fprintf(fout, "<?xml version=\"1.0\" standalone=\"no\"?>\n");
  fprintf(fout, "<!DOCTYPE svg PUBLIC \"-//W3C//DTD SVG 20010904//EN\"\n");
  fprintf(fout, " \"http://www.w3.org/TR/2001/REC-SVG-20010904/DTD/svg10.dtd\">\n");

  /* set bounding box and namespace */
  fprintf(fout, "<svg version=\"1.0\" xmlns=\"http://www.w3.org/2000/svg\"\n");
  fprintf(fout, " width=\"%dpt\" height=\"%dpt\" viewBox=\"0 0 %d %d\"\n", 
	  bboxx, bboxy, bboxx, bboxy);
  fprintf(fout, " preserveAspectRatio=\"xMidYMid meet\">\n");

  /* metadata: creator */
  fprintf(fout, "<metadata>\n");
  fprintf(fout, "Created by "NAME" "VERSION", written by Peter Selinger 2001-2003\n");
  fprintf(fout, "</metadata>\n");

  /* use a "group" tag to establish coordinate system and style */
  fprintf(fout, "<g transform=\"");
  if (origx != 0 || origy != 0) {
    fprintf(fout, "translate(%.0f,%.0f) ", origx, origy);
  }
  if (info.angle != 0) {
    fprintf(fout, "rotate(%.2f) ", -info.angle);
  }
  fprintf(fout, "scale(%f,%f)", scalex, scaley);
  fprintf(fout, "\"\n");
  fprintf(fout, "fill=\"#%06x\" stroke=\"none\">\n", info.color);

  /* write paths */
  list_forall (p, plist) {
    if (p->sign == '+') {
      fprintf(fout, "<path d=\"");
      column = 9;
      newline = 1;
      lastop = 0;
    }
    if (info.debug == 1) {
      svg_jaggy_path(fout, p->pt, p->len, (p->sign == '+'));
    } else if (info.opticurve) {
      svg_path(fout, p->ocurve, p->om, (p->sign == '+'));
    } else {
      svg_path(fout, p->curve, p->m, (p->sign == '+'));
    }
    if (p->next == NULL || p->next->sign == '+') {
      fprintf(fout, "\"/>\n");
    }
  }

  /* write footer */
  fprintf(fout, "</g>\n");
  fprintf(fout, "</svg>\n");

  return 0;
}

