/* Copyright (C) 2001-2003 Peter Selinger.
   This file is part of potrace. It is free software and it is covered
   by the GNU general public license. See the file COPYING for details. */

/* $Id: path.c,v 1.12 2003/08/17 00:39:43 selinger Exp $ */

#include <stdio.h>
#include <stdlib.h>
#include <errno.h>

#include "pbm.h"
#include "curve.h"
#include "main.h"
#include "lists.h"

/* ---------------------------------------------------------------------- */
/* auxiliary functions */

/* return a "random" value which deterministically depends on x,y */
static int detrand(int x, int y) {
  srand(x+123456789*y);
  return rand();
}

/* return the "majority" value of bitmap bm at intersection (x,y). We
   assume that the bitmap is balanced at "radius" 1.  */
static int majority(bitmap_t *bm, int x, int y) {
  int i, a, ct;

  for (i=2; i<5; i++) { /* check at "radius" i */
    ct = 0;
    for (a=-i+1; a<=i-1; a++) {
      ct += BM_GET(bm, x+a, y+i-1) ? 1 : -1;
      ct += BM_GET(bm, x+i-1, y+a-1) ? 1 : -1;
      ct += BM_GET(bm, x+a-1, y-i) ? 1 : -1;
      ct += BM_GET(bm, x-i, y+a) ? 1 : -1;
    }
    if (ct>0) {
      return 1;
    } else if (ct<0) {
      return 0;
    }
  }
  return 0;
}

/* ---------------------------------------------------------------------- */
/* decompose image into paths */

/* a path is represented as an array of points, which are thought to
   lie on the corners of pixels (not on their centers). The path point
   (x,y) is the lower left corner of the pixel (x,y). Paths are
   represented by the len/pt components of a path_t objects (which
   also store other information about the path) */

/* xor the given pixmap with the interior of the given path. Note: the
   path must be within the dimensions of the pixmap. */
static void xor_path(bitmap_t *bm, path_t *p) {
  int xa, x, y, i, j, k, y1;

  if (p->len <= 0) {  /* a path of length 0 is silly, but legal */
    return;
  }

  y1 = p->pt[p->len-1].y;

  xa = p->pt[0].x & -WORDBITS;  /* x0 rounded down to a multiple of WORDBITS */
  for (k=0; k<p->len; k++) {
    x = p->pt[k].x;
    y = p->pt[k].y;

    if (y != y1) {
      /* efficiently invert the rectangle [x,xa] x [y,y1] */
      int xhi = x & -WORDBITS;
      int xlo = x % WORDBITS;
      
      j = (y1+y-1)/2;
      if (xhi<xa) {
	for (i = xhi; i < xa; i+=WORDBITS) {
	  *bm_index(bm, i, j) ^= -1;
	}
      } else {
	for (i = xa; i < xhi; i+=WORDBITS) {
	  *bm_index(bm, i, j) ^= -1;
	}
      }
      /* note: the following "if" is needed b/c x86 treats 
	 a<<b as a<<(b&31)! (It took me hours to find this bug). */
      if (xlo) {
	*bm_index(bm, xhi, j) ^= (-1 << (WORDBITS - xlo));
      }
      y1 = y;
    }
  }
}

/* compute a path in the given pixmap, separating black from white.
   Start path at the point (x0,x1), which must be an upper left corner
   of the path. Also compute the area enclosed by the path. p must
   point to a valid path_t object, whose fields will be filled in. In
   yminp, xminp, xmaxp, return minimum/maximum y/x coordinate
   encountered. */
int findpath(bitmap_t *bm, int x0, int y0, 
	     path_t *p, int *yminp, int *xminp, int *xmaxp) {
  int x, y, dirx, diry, len, size, area;
  int c, d, tmp, ymin, xmin, xmax;
  point_t *pt, *pt1;
  int errno_save;

  x = x0;
  y = y0;
  dirx = 0;
  diry = -1;

  ymin = y0;
  xmin = xmax = x0;

  len = size = 0;
  pt = NULL;
  area = 0;
  
  while (1) {
    /* add point to path */
    if (len>=size) {
      size+=100;
      pt1 = realloc(pt, size * sizeof(point_t));
      if (!pt1) {
	goto error;
      }
      pt = pt1;
    }
    pt[len].x = x;
    pt[len].y = y;
    len++;
    
    /* move to next point */
    x += dirx;
    y += diry;
    if (y < ymin) {
      ymin = y;
    }
    if (x < xmin) {
      xmin = x;
    }
    if (x > xmax) {
      xmax = x;
    }
    area += x*diry;
    
    /* path complete? */
    if (x==x0 && y==y0) {
      break;
    }
    
    /* determine next direction */
    c = BM_GET(bm, x + (dirx+diry-1)/2, y + (diry-dirx-1)/2);
    d = BM_GET(bm, x + (dirx-diry-1)/2, y + (diry+dirx-1)/2);
    
    if (c && !d) {               /* ambiguous turn */
      if (info.turnpolicy == POLICY_RIGHT
	  || (info.turnpolicy == POLICY_BLACK && p->sign == '+')
	  || (info.turnpolicy == POLICY_WHITE && p->sign == '-')
	  || (info.turnpolicy == POLICY_RANDOM && (detrand(x,y) & 1))
	  || (info.turnpolicy == POLICY_MAJORITY && majority(bm, x, y))
	  || (info.turnpolicy == POLICY_MINORITY && !majority(bm, x, y))) {
	tmp = dirx;              /* right turn */
	dirx = diry;
	diry = -tmp;
      } else {
	tmp = dirx;              /* left turn */
	dirx = -diry;
	diry = tmp;
      }
    } else if (c) {              /* right turn */
      tmp = dirx;
      dirx = diry;
      diry = -tmp;
    } else if (!d) {             /* left turn */
      tmp = dirx;
      dirx = -diry;
      diry = tmp;
    }
  } /* while this path */

  p->pt = pt;
  p->len = len;
  p->area = area;

  *yminp = ymin;
  *xminp = xmin;
  *xmaxp = xmax;
  return 0;
 
 error:
  errno_save = errno;
  free(pt);
  errno = errno_save;
  return 1; 
}


/* Decompose the given bitmap into paths. Returns a linked list of
   path_t objects with the fields len, pt, area, sign filled
   in. Returns 0 on success with plistp set, or -1 on error with errno
   set. */

int bm_to_pathlist(bitmap_t *bm, path_t **plistp) {
  int x,y;
  bitmap_t *bm1;
  int yscan, yscan1;
  path_t *p;
  path_t *plist = NULL;  /* linked list of path objects */
  path_t **hook = &plist;  /* used to speed up appending to linked list */
  int r;
  int errno_save;
  int sign;
  int ymin, xmin, xmax;
  int mask_y0 = 0;  /* the mask_* variables hold "bounding box" in mask mode */
  int mask_x0 = 0;
  int mask_x1 = 0;

  /* be sure the byte padding on the right is set to 0, as the fast
     pixel search below relies on it */
  if (bm->w % WORDBITS != 0) {
    word mask = -1 << (WORDBITS - (bm->w % WORDBITS));
    for (y=0; y<bm->h; y++) {
      *bm_index(bm, bm->w, y) &= mask;
    }
  }

  bm1 = bm_new(bm->w, bm->h);
  if (!bm1) {
    goto error;
  }
  bm_clear(bm1, 0);

  yscan=bm->h-1;
  yscan1 = yscan;

  sign = '+';

  while (1) {   /* each iteration produces one path */
    /* if sign=='+', search for the next path. If sign=='-', search
       for the next path that is contained within bm1. */

    /* find the first pixel that is set */

    if (sign == '+') {
      for (; yscan>=0; yscan--) {
	for (x=0; x<bm->w; x+=WORDBITS) {
	  if (*bm_index(bm, x, yscan)) {
	    while (!BM_GET(bm, x, yscan)) {
	      x++;
	    }
	    y = yscan+1;
	    yscan1 = yscan;
	    goto found;
	  }
	}
      }
      /* not found, done */
      break; 
    } else {
      for (; yscan1>=mask_y0; yscan1--) {
	for (x=mask_x0; x<mask_x1; x+=WORDBITS) {
	  if (*bm_index(bm, x, yscan1) & *bm_index(bm1, x, yscan1)) {
	    while (!(BM_GET(bm, x, yscan1) && BM_GET(bm1, x, yscan1))) {
	      x++;
	    }
	    y = yscan1+1;
	    goto found;
	  }
	}
      }
      /* not found, revert to sign == '+' */
      sign = '+';
      /* clear only the part of the mask that was used */
      for (y=yscan; y>=mask_y0; y--) {  
	for (x=mask_x0; x<mask_x1; x+=WORDBITS) {
	  *bm_index(bm1, x, y) = 0;
	}
      }
      continue;
    }      

  found:
    /* allocate new path object */
    p = path_new();
    if (!p) {
      goto error;
    }

    p->sign = sign;

    /* calculate the path */
    r = findpath(bm, x, y, p, &ymin, &xmin, &xmax);
    if (r) {
      goto error;
    }

    if (sign == '+') {
      mask_y0 = ymin;
      mask_x0 = (xmin / WORDBITS) * WORDBITS;
      mask_x1 = ((xmax-1) / WORDBITS + 1) * WORDBITS;
    }

    /* update buffered image */
    xor_path(bm, p);
    /* update mask */
    xor_path(bm1, p);

    /* if it's a turd, eliminate it, else append it to the list */
    if (p->area <= info.turdsize) {
      path_free(p);
    } else {
      list_insert_athook(p, hook);
      hook=&(*hook)->next;
    }

    sign = '-';
  } /* while there are paths */
  
  bm_free(bm1);
  *plistp = plist;
  return 0;

 error:
  errno_save = errno;
  bm_free(bm1);
  list_forall_unlink(p, plist) {
    path_free(p);
  }
  errno = errno_save;
  return -1;
}
