/* Copyright (C) 2001-2003 Peter Selinger.
   This file is part of potrace. It is free software and it is covered
   by the GNU general public license. See the file COPYING for details. */

/* $Id: pgm.c,v 1.6 2003/08/18 02:48:45 selinger Exp $ */

/* Routines for manipulating greymaps, including reading pgm files. We
   only deal with greymaps of depth 8 bits. */

#include <stdlib.h>
#include <errno.h>
#include <string.h>
#include <math.h>

#include "pgm.h"

#define mod(a,n) ((a)>=(n) ? (a)%(n) : (a)>=0 ? (a) : (n)-1-(-1-(a))%(n))

/* ---------------------------------------------------------------------- */
/* basic greymap routines */

/* return new un-initialized greymap. NULL with errno on error */

greymap_t *gm_new(int w, int h) {
  greymap_t *gm;
  int errno_save;

  gm = (greymap_t *) malloc(sizeof(greymap_t));
  if (!gm) {
    return NULL;
  }
  gm->w = w;
  gm->h = h;
  gm->map = (signed short int *) malloc(w*h*sizeof(signed short int));
  if (!gm->map) {
    errno_save = errno;
    free(gm);
    errno = errno_save;
    return NULL;
  }
  return gm;
}

/* free the given greymap */
void gm_free(greymap_t *gm) {
  if (gm) {
    free(gm->map);
  }
  free(gm);
}

/* duplicate the given greymap. Return NULL on error with errno set. */
greymap_t *gm_dup(greymap_t *gm) {
  greymap_t *gm1 = gm_new(gm->w, gm->h);
  if (!gm1) {
    return NULL;
  }
  memcpy(gm1->map, gm->map, gm->w*gm->h*2);
  return gm1;
}

/* clear the given greymap to color b. */
void gm_clear(greymap_t *gm, int b) {
  int i;

  if (b==0) {
    memset(gm->map, 0, gm->w*gm->h*2);
  } else {
    for (i=0; i<gm->w*gm->h; i++) {
      gm->map[i] = b;
    }
  }    
}

/* ---------------------------------------------------------------------- */
/* routines for reading pgm streams */

/* skip whitespace and comments, then read a non-negative decimal
   number from a stream. Return -1 on EOF. Tolerate other errors (skip
   bad characters). Also read the immediately following character, or
   to the end of the line if next character is a comment character. */

static int readnum(FILE *f) {
  int c;
  int acc;

  /* skip whitespace and comments */
  while (1) {
    c = fgetc(f);
    if (c=='#') {
      while (1) {
	c = fgetc(f);
	if (c=='\n' || c==EOF) {
	  break;
	}
      }
    }
    if (c==EOF) {
      return -1;
    }
    if (c>='0' && c<='9') {
      break;
    }
  }

  /* first digit is already in c */
  acc = c-'0';
  while (1) {
    c = fgetc(f);
    if (c=='#') {
      while (1) {
	c = fgetc(f);
	if (c=='\n' || c==EOF) {
	  break;
	}
      }
    }
    if (c<'0' || c>'9') {
      break;
    }
    acc *= 10;
    acc += c-'0';
  }
  return acc;
}

/* read a pgm stream: either P2 or P5 format. See pgm(5). Return
   bitmap in *gmp. Return 0 on success, -1 on error with errno set, -2
   on bad file format, and 1 on premature end of file. If the return
   value is >=0, *gmp is valid. */

int gm_readpgm(FILE *f, greymap_t **gmp) {
  int magic[2];
  greymap_t *gm;
  int x, y, b;
  int max;
  int w, h;

  /* read header */
  magic[0] = fgetc(f);
  if (magic[0] != 'P') {
    return -2;
  }

  magic[1] = fgetc(f);
  if (magic[1] != '2' && magic[1] != '5') {
    return -2;
  }

  w = readnum(f);
  if (w<0) {
    return -2;
  }

  h = readnum(f);
  if (h<0) {
    return -2;
  }

  max = readnum(f);
  if (max<0) {
    return -2;
  }

  /* allocate greymap */
  gm = gm_new(w, h);
  if (!gm) {
    return -1;
  }

  /* zero it out */
  gm_clear(gm, 0);

  switch (magic[1]) {
  case '2':
  default:
    /* read P2 format */
    
    for (y=h-1; y>=0; y--) {
      for (x=0; x<w; x++) {
	b = readnum(f);
	if (b<0) {
	  goto eof;
	}
	GM_UPUT(gm, x, y, (b*255/max));
      }
    }
    break;

  case '5':
    /* read P5 format */
    
    for (y=h-1; y>=0; y--) {
      for (x=0; x<w; x++) {
	b = fgetc(f);
	if (b==EOF) {
	  goto eof;
	}
	GM_UPUT(gm, x, y, (b*255/max));
      }
    }
    break;
  }

  *gmp = gm;
  return 0;

 eof:
  *gmp = gm;
  return 1;
}

/* write a pgm stream, either P2 or (if raw != 0) P5 format. Include
   one-line comment if non-NULL. Mode determines how out-of-range
   color values are converted. Gamma is the desired gamma correction,
   if any (set to 2.2 if the image is to look optimal on a CRT monitor,
   2.8 for LCD). Set to 1.0 for no gamma correction */

int gm_writepgm(FILE *f, greymap_t *gm, char *comment, int raw, int mode, double gamma) {
  int x, y, v;
  int gammatable[256];
  
  /* prepare gamma correction lookup table */
  if (gamma != 1.0) {
    for (v=0; v<256; v++) {
      gammatable[v] = rint(255 * exp(log(v/255.0)/gamma));
    }
  } else {
    for (v=0; v<256; v++) {
      gammatable[v] = v;
    }
  }  

  fprintf(f, raw ? "P5\n" : "P2\n");
  if (comment && *comment) {
    fprintf(f, "# %s\n", comment);
  }
  fprintf(f, "%d %d 255\n", gm->w, gm->h);
  for (y=gm->h-1; y>=0; y--) {
    for (x=0; x<gm->w; x++) {
      v = GM_UGET(gm, x, y);
      if (mode == GM_MODE_NONZERO) {
	if (v > 255) {
	  v = 510 - v;
	}
	if (v < 0) {
	  v = 0;
	}
      } else if (mode == GM_MODE_ODD) {
	v = mod(v, 510);
	if (v > 255) {
	  v = 510 - v;
	}
      } else if (mode == GM_MODE_POSITIVE) {
	if (v < 0) {
	  v = 0;
	} else if (v > 255) {
	  v = 255;
	}
      } else if (mode == GM_MODE_NEGATIVE) {
	v = 510 - v;
	if (v < 0) {
	  v = 0;
	} else if (v > 255) {
	  v = 255;
	}
      }
      v = gammatable[v];
      
      if (raw) {
	fputc(v, f);
      } else {
	fprintf(f, x == gm->w-1 ? "%d\n" : "%d ", v);
      }	  
    }
  }
  return 0;
}

/* ---------------------------------------------------------------------- */
/* output - for primitive debugging purposes only! */

/* print greymap to screen */
int gm_print(FILE *f, greymap_t *gm) {
  int x, y;
  int xx, yy;
  int d, t;
  int sw, sh;

  sw = gm->w < 79 ? gm->w : 79;
  sh = gm->w < 79 ? gm->h : gm->h*sw*44/(79*gm->w);

  for (yy=sh-1; yy>=0; yy--) {
    for (xx=0; xx<sw; xx++) {
      d=0;
      t=0;
      for (x=xx*gm->w/sw; x<(xx+1)*gm->w/sw; x++) {
	for (y=yy*gm->h/sh; y<(yy+1)*gm->h/sh; y++) {
	  d += GM_GET(gm, x, y);
	  t += 256;
	}
      }
      fputc("*#=- "[5*d/t], f);
    }
    fputc('\n', f);
  }
  return 0;
}
